#define IMPORT_CORE
#include "core/HashMap.h"

#include "core/Utility.h"

static bool isInvalidKey(const HashMap* m, const void* key) {
    const char* c = key;
    for(size_t i = m->keySize; i > 0; i--) {
        if(*(c++) != 0) {
            return false;
        }
    }
    return true;
}

static void* getKey(const HashMap* m, size_t index) {
    return (char*)m->keys + m->keySize * index;
}

static void* getValue(const HashMap* m, size_t index) {
    return (char*)m->values + m->valueSize * index;
}

static size_t searchSlot(HashMap* m, const void* key) {
    size_t rehashFactor = 2;
    while(true) {
        rehashHashMap(m, m->entries * rehashFactor + 1);
        if(isInvalidKey(m, key)) {
            return m->capacity - 1;
        }
        size_t baseHash = m->hasher(key, m->keySize) * 514685581u;
        size_t end = m->capacity - 2;
        // rehash on bad clustering
        for(size_t i = 0; i <= 5; i++) {
            size_t hash = (baseHash + i) & end;
            void* keyEntry = getKey(m, hash);
            if(isInvalidKey(m, keyEntry) ||
               m->equal(keyEntry, key, m->keySize)) {
                return hash;
            }
        }
        rehashFactor *= 2;
    }
}

#define SEARCH_VALUE(type, Type)                                               \
    static void* searchValue##Type(const HashMap* m, const void* rawKey) {     \
        type key = *(const type*)rawKey;                                       \
        if(m->capacity != 0) {                                                 \
            if(key == 0) {                                                     \
                size_t i = m->capacity - 1;                                    \
                return ((type*)m->keys)[i] == 0 ? getValue(m, i) : nullptr;    \
            }                                                                  \
            size_t baseHash = ((size_t)key) * 514685581u;                      \
            size_t end = m->capacity - 2;                                      \
            for(size_t i = 0; i <= end; i++) {                                 \
                size_t hash = (baseHash + i) & end;                            \
                type keyEntry = *(type*)getKey(m, hash);                       \
                if(keyEntry == key) {                                          \
                    return getValue(m, hash);                                  \
                } else if(keyEntry == 0) {                                     \
                    return nullptr;                                            \
                }                                                              \
            }                                                                  \
        }                                                                      \
        return nullptr;                                                        \
    }
SEARCH_VALUE(u32, U32)
SEARCH_VALUE(u64, U64)

static void* searchValue(const HashMap* m, const void* key) {
    if(m->capacity != 0) {
        if(isInvalidKey(m, key)) {
            size_t i = m->capacity - 1;
            return isInvalidKey(m, getKey(m, i)) ? getValue(m, i) : nullptr;
        }
        size_t baseHash = m->hasher(key, m->keySize) * 514685581u;
        size_t end = m->capacity - 2;
        for(size_t i = 0; i <= end; i++) {
            size_t hash = (baseHash + i) & end;
            void* keyEntry = getKey(m, hash);
            if(m->equal(keyEntry, key, m->keySize)) {
                return getValue(m, hash);
            } else if(isInvalidKey(m, keyEntry)) {
                return nullptr;
            }
        }
    }
    return nullptr;
}

static size_t roundUp2(size_t n) {
    size_t w = 1;
    while(w < n) {
        w *= 2;
    }
    return w;
}

void initHashMap(HashMap* m, size_t keySize, size_t valueSize) {
    initHashMapF(m, keySize, valueSize, hashKey, compareKeys);
}

void initHashMapF(HashMap* m, size_t keySize, size_t valueSize, Hasher hasherF,
                  Equal equal) {
    SearchValue search = searchValue;
    if(hasherF == hashKey && equal == compareKeys) {
        switch(keySize) {
            case sizeof(u32): search = searchValueU32; break;
            case sizeof(u64): search = searchValueU64; break;
        }
    }
    m->keys = nullptr;
    m->values = nullptr;
    m->keySize = keySize;
    m->valueSize = valueSize;
    m->capacity = 0;
    m->entries = 0;
    m->hasher = hasherF;
    m->equal = equal;
    m->search = search;
}

void destroyHashMap(HashMap* m) {
    cFree(m->keys);
    cFree(m->values);
    *m = (HashMap){0};
}

void rehashHashMap(HashMap* m, size_t minCapacity) {
    if(minCapacity <= m->capacity) {
        return;
    }
    size_t l = roundUp2(maxSize(minCapacity, 8lu)) + 1;
    HashMap map;
    initHashMapF(&map, m->keySize, m->valueSize, m->hasher, m->equal);
    size_t keyBytes = l * m->keySize;
    map.keys = cAllocate(keyBytes);
    memset(map.keys, 0, keyBytes);
    memset(getKey(&map, l - 1), 1, m->keySize);
    map.values = cAllocate(l * m->valueSize);
    map.capacity = l;

    size_t length = m->capacity;
    if(length > 0) {
        length--;
        for(size_t i = 0; i < length; i++) {
            void* keyEntry = getKey(m, i);
            if(!isInvalidKey(m, keyEntry)) {
                putHashMapPair(&map, keyEntry, getValue(m, i));
            }
        }
        void* keyEntry = getKey(m, length);
        if(isInvalidKey(m, keyEntry)) {
            putHashMapPair(&map, keyEntry, getValue(m, length));
        }
    }
    swap(&map, m);
    destroyHashMap(&map);
}

void* putHashMapPair(HashMap* m, const void* key, const void* value) {
    size_t index = searchSlot(m, key);
    void* keyEntry = getKey(m, index);
    if(m->equal(keyEntry, key, m->keySize)) {
        void* v = getValue(m, index);
        memcpy(v, value, m->valueSize);
        return v;
    }
    void* v = getValue(m, index);
    memcpy(v, value, m->valueSize);
    m->entries++;
    memcpy(keyEntry, key, m->keySize);
    return v;
}

void* searchHashMapKey(const HashMap* m, const void* key) {
    return m->search(m, key);
}

void clearHashMap(HashMap* m) {
    if(m->keys == nullptr) {
        return;
    }
    m->entries = 0;
    memset(m->keys, 0, m->capacity * m->keySize);
    memset(getKey(m, m->capacity - 1), 1, m->keySize);
}

bool removeHashMapKey(HashMap* m, const void* key) {
    // ToDo: This is a very slow remove
    HashMap n;
    initHashMapF(&n, m->keySize, m->valueSize, m->hasher, m->equal);
    HashMapIterator i;
    initHashMapIterator(&i, m);
    bool r = false;
    while(hasNextHashMapNode(&i)) {
        HashMapNode* node = nextHashMapNode(&i);
        if(!m->equal(key, node->key, n.keySize)) {
            putHashMapPair(&n, node->key, node->value);
        } else {
            r = true;
        }
    }
    swap(&n, m);
    destroyHashMap(&n);
    return r;
}

size_t hashString(const void* key, size_t n) {
    if(n != sizeof(char*)) {
        return 0;
    }
    const char* s = *(const char* const*)key;
    size_t h = 0;
    while(*s != '\0') {
        h = 2120251889lu * h + (size_t)(*(s++));
    }
    return h;
}

size_t hashKey(const void* key, size_t n) {
    size_t h = 0;
    if(n <= sizeof(size_t)) {
        memcpy(&h, key, n);
        return h;
    }
    const char* cKey = key;
    while(n > 0) {
        size_t m = 0;
        if(n >= sizeof(size_t)) {
            memcpy(&m, cKey, sizeof(size_t));
            cKey += sizeof(size_t);
            n -= sizeof(size_t);
        } else {
            memcpy(&m, cKey, n);
            cKey += n;
            n = 0;
        }
        h ^= m;
    }
    return h;
}

bool compareKeys(const void* keyA, const void* keyB, size_t n) {
    return memcmp(keyA, keyB, n) == 0;
}

void initHashMapIterator(HashMapIterator* mi, const HashMap* m) {
    mi->key = m->keys;
    mi->endKey = mi->key + m->capacity * m->keySize;
    mi->value = m->values;
    mi->endValue = mi->value + m->capacity * m->valueSize;
    mi->map = m;
}

bool hasNextHashMapNode(HashMapIterator* mi) {
    while(mi->key != mi->endKey) {
        char* nextKey = mi->key + mi->map->keySize;
        if(isInvalidKey(mi->map, mi->key) == (nextKey == mi->endKey)) {
            break;
        }
        mi->key = nextKey;
        mi->value += mi->map->valueSize;
    }
    return mi->key != mi->endKey;
}

HashMapNode* nextHashMapNode(HashMapIterator* mi) {
    mi->node.key = mi->key;
    mi->node.value = mi->value;
    mi->key += mi->map->keySize;
    mi->value += mi->map->valueSize;
    return &mi->node;
}
